# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for Google Compute Engine Load Balancer Driver
"""
import sys
import unittest

from libcloud.test.secrets import GCE_PARAMS, GCE_KEYWORD_PARAMS
from libcloud.common.google import GoogleBaseAuthConnection
from libcloud.compute.drivers.gce import GCENodeDriver
from libcloud.test.compute.test_gce import GCEMockHttp
from libcloud.test.common.test_google import GoogleTestCase, GoogleAuthMockHttp
from libcloud.loadbalancer.drivers.gce import GCELBDriver


class GCELoadBalancerTest(GoogleTestCase):
    datacenter = "us-central1-a"

    def setUp(self):
        GCEMockHttp.test = self
        GCELBDriver.connectionCls.conn_class = GCEMockHttp
        GCENodeDriver.connectionCls.conn_class = GCEMockHttp
        GoogleBaseAuthConnection.conn_class = GoogleAuthMockHttp
        GCEMockHttp.type = None
        kwargs = GCE_KEYWORD_PARAMS.copy()
        kwargs["auth_type"] = "IA"
        kwargs["datacenter"] = self.datacenter
        self.driver = GCELBDriver(*GCE_PARAMS, **kwargs)

    def test_get_node_from_ip(self):
        ip = "23.236.58.15"
        expected_name = "node-name"
        node = self.driver._get_node_from_ip(ip)
        self.assertEqual(node.name, expected_name)

        dummy_ip = "8.8.8.8"
        node = self.driver._get_node_from_ip(dummy_ip)
        self.assertTrue(node is None)

    def test_list_protocols(self):
        expected_protocols = ["TCP", "UDP"]
        protocols = self.driver.list_protocols()
        self.assertEqual(protocols, expected_protocols)

    def test_list_balancers(self):
        balancers = self.driver.list_balancers()
        balancers_all = self.driver.list_balancers(ex_region="all")
        balancer_name = "lcforwardingrule"
        self.assertEqual(len(balancers), 2)
        self.assertEqual(len(balancers_all), 2)
        self.assertEqual(balancers[0].name, balancer_name)

    def test_create_balancer(self):
        balancer_name = "libcloud-lb-demo-lb"
        tp_name = "%s-tp" % (balancer_name)
        port = "80"
        protocol = "tcp"
        algorithm = None
        node0 = self.driver.gce.ex_get_node("libcloud-lb-demo-www-000", "us-central1-b")
        node1 = self.driver.gce.ex_get_node("libcloud-lb-demo-www-001", "us-central1-b")
        members = [node0, node1]
        balancer = self.driver.create_balancer(balancer_name, port, protocol, algorithm, members)
        self.assertEqual(balancer.name, balancer_name)
        self.assertEqual(balancer.extra["targetpool"].name, tp_name)
        self.assertEqual(len(balancer.list_members()), 3)

    def test_destory_balancer(self):
        balancer_name = "lcforwardingrule"
        balancer = self.driver.get_balancer(balancer_name)
        destroyed = balancer.destroy()
        self.assertTrue(destroyed)

    def test_get_balancer(self):
        balancer_name = "lcforwardingrule"
        tp_name = "lctargetpool"
        balancer_ip = "173.255.119.224"
        balancer = self.driver.get_balancer(balancer_name)
        self.assertEqual(balancer.name, balancer_name)
        self.assertEqual(balancer.extra["forwarding_rule"].name, balancer_name)
        self.assertEqual(balancer.ip, balancer_ip)
        self.assertEqual(balancer.extra["targetpool"].name, tp_name)

    def test_attach_compute_node(self):
        node = self.driver.gce.ex_get_node("libcloud-lb-demo-www-001", "us-central1-b")
        balancer = self.driver.get_balancer("lcforwardingrule")
        member = self.driver._node_to_member(node, balancer)
        # Detach member first
        balancer.detach_member(member)
        self.assertEqual(len(balancer.list_members()), 1)
        # Attach Node
        balancer.attach_compute_node(node)
        self.assertEqual(len(balancer.list_members()), 2)

    def test_detach_attach_member(self):
        node = self.driver.gce.ex_get_node("libcloud-lb-demo-www-001", "us-central1-b")
        balancer = self.driver.get_balancer("lcforwardingrule")
        member = self.driver._node_to_member(node, balancer)

        # Check that balancer has 2 members
        self.assertEqual(len(balancer.list_members()), 2)

        # Remove a member and check that it now has 1 member
        balancer.detach_member(member)
        self.assertEqual(len(balancer.list_members()), 1)

        # Reattach member and check that it has 2 members again
        balancer.attach_member(member)
        self.assertEqual(len(balancer.list_members()), 2)

    def test_balancer_list_members(self):
        balancer = self.driver.get_balancer("lcforwardingrule")
        members = balancer.list_members()
        self.assertEqual(len(members), 2)
        member_ips = [m.ip for m in members]
        self.assertTrue("23.236.58.15" in member_ips)

    def test_ex_create_healthcheck(self):
        healthcheck_name = "lchealthcheck"
        kwargs = {
            "host": "lchost",
            "path": "/lc",
            "port": 8000,
            "interval": 10,
            "timeout": 10,
            "unhealthy_threshold": 4,
            "healthy_threshold": 3,
        }
        hc = self.driver.ex_create_healthcheck(healthcheck_name, **kwargs)
        self.assertEqual(hc.name, healthcheck_name)
        self.assertEqual(hc.path, "/lc")
        self.assertEqual(hc.port, 8000)
        self.assertEqual(hc.interval, 10)

    def test_ex_list_healthchecks(self):
        healthchecks = self.driver.ex_list_healthchecks()
        self.assertEqual(len(healthchecks), 3)
        self.assertEqual(healthchecks[0].name, "basic-check")

    def test_ex_balancer_detach_attach_healthcheck(self):
        healthcheck = self.driver.gce.ex_get_healthcheck("libcloud-lb-demo-healthcheck")
        balancer = self.driver.get_balancer("lcforwardingrule")

        healthchecks = self.driver.ex_balancer_list_healthchecks(balancer)
        self.assertEqual(len(healthchecks), 1)
        # Detach Healthcheck
        detach_healthcheck = self.driver.ex_balancer_detach_healthcheck(balancer, healthcheck)
        self.assertTrue(detach_healthcheck)
        healthchecks = self.driver.ex_balancer_list_healthchecks(balancer)
        self.assertEqual(len(healthchecks), 0)

        # Reattach Healthcheck
        attach_healthcheck = self.driver.ex_balancer_attach_healthcheck(balancer, healthcheck)
        self.assertTrue(attach_healthcheck)
        healthchecks = self.driver.ex_balancer_list_healthchecks(balancer)
        self.assertEqual(len(healthchecks), 1)

    def test_ex_balancer_list_healthchecks(self):
        balancer = self.driver.get_balancer("lcforwardingrule")
        healthchecks = self.driver.ex_balancer_list_healthchecks(balancer)
        self.assertEqual(healthchecks[0].name, "libcloud-lb-demo-healthcheck")

    def test_node_to_member(self):
        node = self.driver.gce.ex_get_node("libcloud-lb-demo-www-001", "us-central1-b")
        balancer = self.driver.get_balancer("lcforwardingrule")
        member = self.driver._node_to_member(node, balancer)
        self.assertEqual(member.ip, node.public_ips[0])
        self.assertEqual(member.id, node.name)
        self.assertEqual(member.port, balancer.port)

    def test_node_to_member_no_pub_ip(self):
        node = self.driver.gce.ex_get_node("libcloud-lb-nopubip-001", "us-central1-b")
        balancer = self.driver.get_balancer("lcforwardingrule")
        member = self.driver._node_to_member(node, balancer)
        self.assertIsNone(member.ip)

    def test_forwarding_rule_to_loadbalancer(self):
        fwr = self.driver.gce.ex_get_forwarding_rule("lcforwardingrule")
        balancer = self.driver._forwarding_rule_to_loadbalancer(fwr)
        self.assertEqual(fwr.name, balancer.name)
        self.assertEqual(fwr.address, balancer.ip)
        self.assertEqual(fwr.extra["portRange"], balancer.port)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64

from libcloud.utils.py3 import b
from libcloud.common.base import JsonResponse, ConnectionUserAndKey
from libcloud.common.types import ProviderError

__all__ = [
    "API_HOST",
    "LiquidWebException",
    "LiquidWebResponse",
    "LiquidWebConnection",
]


# Endpoint for liquidweb api.
API_HOST = "api.stormondemand.com"


class LiquidWebException(ProviderError):
    """The base class for other Liquidweb exceptions"""

    def __init__(self, value, http_code, extra=None):
        """
        :param value: message contained in error
        :type  value: ``str``

        :param http_code: error code
        :type  http_code: ``int``

        :param extra: extra fields specific to error type
        :type  extra: ``list``
        """
        self.extra = extra
        super().__init__(value, http_code, driver=None)

    def __str__(self):
        return "{}  {}".format(self.http_code, self.value)

    def __repr__(self):
        return "LiquidWebException {} {}".format(self.http_code, self.value)


class APIException(LiquidWebException):
    def __init__(self, error_class, full_msg, http_code, extra=None):
        self.error_class = error_class
        super().__init__(full_msg, http_code, extra=extra)

    def __str__(self):
        return "{}: {}".format(self.error_class, self.value)

    def __repr__(self):
        return "{}: {}".format(self.error_class, self.value)


EXCEPTIONS_FIELDS = {
    "LW::Exception::API::Internal": {"fields": []},
    "LW::Exception::API::InvalidEncoding": {"fields": ["encoding"]},
    "LW::Exception::API::InvalidMethod": {"fields": ["method"]},
    "LW::Exception::API::Maintenance": {"fields": []},
    "LW::Exception::API::RateLimit": {"fields": ["account", "ip", "method"]},
    "LW::Exception::Authorization": {"fields": ["username"]},
    "LW::Exception::DNS::NoResponse": {"fields": ["nameservers"]},
    "LW::Exception::DNS::Servfail": {"fields": ["nameservers"]},
    "LW::Exception::Deserialize": {"fields": ["data", "encoding"]},
    "LW::Exception::DuplicateRecord": {"fields": ["field", "input", "statement"]},
    "LW::Exception::Forbidden": {"fields": []},
    "LW::Exception::Incapable": {"fields": ["capability", "thing"]},
    "LW::Exception::Input": {"fields": ["field"]},
    "LW::Exception::Input::Disallowed": {"fields": ["field"]},
    "LW::Exception::Input::Multiple": {"fields": ["errors", "field", "type"]},
    "LW::Exception::Input::NotInRealm": {"fields": ["field", "valid", "value"]},
    "LW::Exception::Input::OutOfBounds": {"fields": ["field", "max", "min", "value"]},
    "LW::Exception::Input::Required": {"fields": ["field", "position"]},
    "LW::Exception::Input::Unknown": {"fields": ["field", "value"]},
    "LW::Exception::Input::Validation": {"fields": ["field", "type", "value"]},
    "LW::Exception::Permission": {"fields": ["account", "identifier"]},
    "LW::Exception::RecordNotFound": {"fields": ["field", "input"]},
    "LW::Exception::RemoteService::Authorization": {"fields": ["url"]},
    "LW::Exception::Resource": {"fields": ["resource"]},
    "LW::Exception::Resource::Insufficient": {"fields": ["available", "requested", "resource"]},
    "LW::Exception::Resource::Unavailable": {"fields": ["resource"]},
    "LW::Exception::Serialize": {"fields": ["data", "encoding"]},
    "LW::Exception::Workflow::Conflict": {"fields": ["conflict", "workflow"]},
}


class LiquidWebResponse(JsonResponse):
    objects = None
    errors = None

    def __init__(self, response, connection):
        self.errors = []
        super().__init__(response=response, connection=connection)
        self.objects, self.errors = self.parse_body_and_errors()
        if self.errors:
            error = self.errors.pop()
            raise self._make_excp(error, self.status)

    def parse_body_and_errors(self):
        data = []
        errors = []
        js = super().parse_body()
        if "items" in js:
            data.append(js["items"])

        if "name" in js:
            data.append(js)

        if "deleted" in js:
            data.append(js["deleted"])

        if "error_class" in js:
            errors.append(js)

        return (data, errors)

    def success(self):
        """
        Returns ``True`` if our request is successful.
        """
        return len(self.errors) == 0

    def _make_excp(self, error, status):
        """
        Raise LiquidWebException.
        """
        exc_type = error.get("error_class")
        message = error.get("full_message")
        try:
            _type = EXCEPTIONS_FIELDS[exc_type]
            fields = _type.get("fields")
            extra = {}
        except KeyError:
            fields = []

        for field in fields:
            extra[field] = error.get(field)
        return APIException(exc_type, message, status, extra=extra)


class LiquidWebConnection(ConnectionUserAndKey):
    host = API_HOST
    responseCls = LiquidWebResponse

    def add_default_headers(self, headers):
        b64string = b("{}:{}".format(self.user_id, self.key))
        encoded = base64.b64encode(b64string).decode("utf-8")
        authorization = "Basic " + encoded

        headers["Authorization"] = authorization
        headers["Content-Type"] = "application/json"

        return headers



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.common.types import LibcloudError
from libcloud.test.storage.test_s3 import S3Tests, S3MockHttp
from libcloud.storage.drivers.auroraobjects import AuroraObjectsStorageDriver


class AuroraObjectsTests(S3Tests, unittest.TestCase):
    driver_type = AuroraObjectsStorageDriver

    def setUp(self):
        super().setUp()

        AuroraObjectsStorageDriver.connectionCls.conn_class = S3MockHttp
        S3MockHttp.type = None
        self.driver = self.create_driver()

    def test_get_object_cdn_url(self):
        self.mock_response_klass.type = "get_object"
        obj = self.driver.get_object(container_name="test2", object_name="test")

        with self.assertRaises(LibcloudError):
            self.driver.get_object_cdn_url(obj)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.types import LibcloudError

__all__ = [
    "Provider",
    "RecordType",
    "ZoneError",
    "ZoneDoesNotExistError",
    "ZoneAlreadyExistsError",
    "RecordError",
    "RecordDoesNotExistError",
    "RecordAlreadyExistsError",
    "OLD_CONSTANT_TO_NEW_MAPPING",
]


class Provider:
    """
    Defines for each of the supported providers

    Non-Dummy drivers are sorted in alphabetical order. Please preserve this
    ordering when adding new drivers.
    """

    DUMMY = "dummy"
    AURORADNS = "auroradns"
    BUDDYNS = "buddyns"
    CLOUDFLARE = "cloudflare"
    DIGITAL_OCEAN = "digitalocean"
    DNSIMPLE = "dnsimple"
    DURABLEDNS = "durabledns"
    GANDI = "gandi"
    GANDI_LIVE = "gandi_live"
    GODADDY = "godaddy"
    GOOGLE = "google"
    LINODE = "linode"
    LIQUIDWEB = "liquidweb"
    LUADNS = "luadns"
    NFSN = "nfsn"
    NSONE = "nsone"
    ONAPP = "onapp"
    POINTDNS = "pointdns"
    POWERDNS = "powerdns"
    RACKSPACE = "rackspace"
    RCODEZERO = "rcodezero"
    ROUTE53 = "route53"
    VULTR = "vultr"
    WORLDWIDEDNS = "worldwidedns"
    ZERIGO = "zerigo"
    ZONOMI = "zonomi"
    DNSPOD = "dnspod"
    # Deprecated
    RACKSPACE_US = "rackspace_us"
    RACKSPACE_UK = "rackspace_uk"


OLD_CONSTANT_TO_NEW_MAPPING = {
    Provider.RACKSPACE_US: Provider.RACKSPACE,
    Provider.RACKSPACE_UK: Provider.RACKSPACE,
}


class RecordType:
    # TODO: Fix all the affected code and tests and use base Type class here
    """
    DNS record type.
    """
    A = "A"
    AAAA = "AAAA"
    AFSDB = "A"
    ALIAS = "ALIAS"
    CERT = "CERT"
    CNAME = "CNAME"
    DNAME = "DNAME"
    DNSKEY = "DNSKEY"
    DS = "DS"
    GEO = "GEO"
    HINFO = "HINFO"
    KEY = "KEY"
    LOC = "LOC"
    MX = "MX"
    NAPTR = "NAPTR"
    NS = "NS"
    NSEC = "NSEC"
    OPENPGPKEY = "OPENPGPKEY"
    PTR = "PTR"
    REDIRECT = "REDIRECT"
    RP = "RP"
    RRSIG = "RRSIG"
    SOA = "SOA"
    SPF = "SPF"
    SRV = "SRV"
    SSHFP = "SSHFP"
    TLSA = "TLSA"
    TXT = "TXT"
    URL = "URL"
    WKS = "WKS"
    CAA = "CAA"


class ZoneError(LibcloudError):
    error_type = "ZoneError"
    kwargs = ("zone_id",)

    def __init__(self, value, driver, zone_id):
        self.zone_id = zone_id
        super().__init__(value=value, driver=driver)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<{} in {}, zone_id={}, value={}>".format(
            self.error_type,
            repr(self.driver),
            self.zone_id,
            self.value,
        )


class ZoneDoesNotExistError(ZoneError):
    error_type = "ZoneDoesNotExistError"


class ZoneAlreadyExistsError(ZoneError):
    error_type = "ZoneAlreadyExistsError"


class RecordError(LibcloudError):
    error_type = "RecordError"

    def __init__(self, value, driver, record_id):
        self.record_id = record_id
        super().__init__(value=value, driver=driver)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<{} in {}, record_id={}, value={}>".format(
            self.error_type,
            repr(self.driver),
            self.record_id,
            self.value,
        )


class RecordDoesNotExistError(RecordError):
    error_type = "RecordDoesNotExistError"


class RecordAlreadyExistsError(RecordError):
    error_type = "RecordAlreadyExistsError"



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import simplejson as json
except ImportError:
    import json

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import (
    Provider,
    RecordType,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
)
from libcloud.common.luadns import LuadnsResponse, LuadnsException, LuadnsConnection

__all__ = ["LuadnsDNSDriver"]


class LuadnsDNSResponse(LuadnsResponse):
    pass


class LuadnsDNSConnection(LuadnsConnection):
    responseCls = LuadnsDNSResponse


class LuadnsDNSDriver(DNSDriver):
    type = Provider.LUADNS
    name = "Luadns"
    website = "https://www.luadns.com"
    connectionCls = LuadnsDNSConnection

    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.MX: "MX",
        RecordType.NS: "NS",
        RecordType.PTR: "PTR",
        RecordType.SOA: "SOA",
        RecordType.SRV: "SRV",
        RecordType.TXT: "TXT",
    }

    def list_zones(self):
        """
        Return a list of zones.

        :return: ``list`` of :class:`Zone`
        """
        action = "/v1/zones"
        response = self.connection.request(action=action, method="GET")
        zones = self._to_zones(response.parse_body())

        return zones

    def get_zone(self, zone_id):
        """
        Return a Zone instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :rtype: :class:`Zone`
        """
        action = "/v1/zones/%s" % zone_id
        try:
            response = self.connection.request(action=action)
        except LuadnsException as e:
            if e.message in ["Zone not found.", "Resource not found."]:
                raise ZoneDoesNotExistError(zone_id=zone_id, value="", driver=self)
            else:
                raise e

        zone = self._to_zone(response.parse_body())

        return zone

    def delete_zone(self, zone):
        """
        Delete a zone.

        Note: This will delete all the records belonging to this zone.

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``
        """
        action = "/v1/zones/%s" % zone.id
        try:
            response = self.connection.request(action=action, method="DELETE")
        except LuadnsException as e:
            if e.message in ["Resource not found.", "Zone not found."]:
                raise ZoneDoesNotExistError(zone_id=zone.id, value="", driver=self)
            else:
                raise e

        return response.status == 200

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        Create a new zone.

        :param domain: Zone domain name (e.g. example.com)
        :type domain: ``str``

        :param type: Zone type (This is not really used. See API docs for extra
                     parameters).
        :type  type: ``str``

        :param ttl: TTL for new records. (This is not really used)
        :type  ttl: ``int``

        :param extra: Extra attributes (driver specific). ('region_support',
                      'zone_data')
        :type extra: ``dict``

        :rtype: :class:`Zone`
        """
        action = "/v1/zones"
        data = json.dumps({"name": domain})
        try:
            response = self.connection.request(action=action, method="POST", data=data)
        except LuadnsException as e:
            if e.message == "Zone '%s' is taken already." % domain:
                raise ZoneAlreadyExistsError(zone_id=domain, value="", driver=self)
            else:
                raise e
        zone = self._to_zone(response.parse_body())

        return zone

    def list_records(self, zone):
        """
        Return a list of records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`Record`
        """
        action = "/v1/zones/%s/records" % zone.id
        response = self.connection.request(action=action)
        records = self._to_records(response.parse_body(), zone=zone)

        return records

    def get_record(self, zone_id, record_id):
        """
        Return a Record instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :param record_id: ID of the required record
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        zone = self.get_zone(zone_id=zone_id)
        action = "/v1/zones/{}/records/{}".format(zone_id, record_id)
        try:
            response = self.connection.request(action=action)
        except LuadnsException as e:
            if e.message == "Record not found.":
                raise RecordDoesNotExistError(record_id=record_id, driver=self, value="")
            else:
                raise e

        record = self._to_record(response.parse_body(), zone=zone)

        return record

    def delete_record(self, record):
        """
        Delete a record.

        :param record: Record to delete.
        :type  record: :class:`Record`

        :rtype: ``bool``
        """
        action = "/v1/zones/{}/records/{}".format(record.zone.id, record.id)
        try:
            response = self.connection.request(action=action, method="DELETE")
        except LuadnsException as e:
            if e.message == "Record not found.":
                raise RecordDoesNotExistError(record_id=record.id, driver=self, value="")
            else:
                raise e

        return response.status == 200

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a record.

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param zone: Zone which the records will be created for.
        :type zone: :class:`Zone`

        :param type: DNS record type ( 'A', 'AAAA', 'CNAME', 'MX', 'NS',
                     'PTR', 'SOA', 'SRV', 'TXT').
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: (optional) Extra attributes ('prio', 'ttl').
        :type  extra: ``dict``

        :rtype: :class:`Record`
        """
        action = "/v1/zones/%s/records" % zone.id
        to_post = {"name": name, "content": data, "type": type, "zone_id": int(zone.id)}
        # ttl is required to create a record for luadns
        # pass it through extra like this: extra={'ttl':ttl}
        if extra is not None:
            to_post.update(extra)
        data = json.dumps(to_post)
        try:
            response = self.connection.request(action=action, method="POST", data=data)
        except LuadnsException as e:
            raise e

        record = self._to_record(response.parse_body(), zone=zone)

        return record

    def _to_zone(self, item):
        common_attr = ["id", "name"]
        extra = {}
        for key in item:
            if key not in common_attr:
                extra[key] = item.get(key)
        zone = Zone(
            domain=item["name"],
            id=item["id"],
            type=None,
            ttl=None,
            driver=self,
            extra=extra,
        )

        return zone

    def _to_zones(self, items):
        zones = []
        for item in items:
            zones.append(self._to_zone(item))

        return zones

    def _to_record(self, item, zone):
        common_attr = ["id", "content", "name", "type"]
        extra = {}
        for key in item:
            if key not in common_attr:
                extra[key] = item.get(key)
        record = Record(
            id=item["id"],
            name=item["name"],
            type=item["type"],
            data=item["content"],
            zone=zone,
            driver=self,
            extra=extra,
        )

        return record

    def _to_records(self, items, zone):
        records = []
        for item in items:
            records.append(self._to_record(item, zone))

        return records
